from .core.AoICalc import MeanAoICalc, MeanPeakAoICalc
from .core.AoIQueueNetwork import AoIQueueNetwork
from .core.AoIAgent import AoIAgent
from .queues.AoIQueueServer import AoIQueueServer
from .queues.LcfsPreemption import LcfsPreemption
from .queues.LgfsPreemption import LgfsPreemption
from .queues.LgfsMultiServerNoPreemption import LgfsMultiServerNoPreemption
from .queues.LgfsMultiServerPreemption import LgfsMultiServerPreemption
from .queues.LgfsMultiServerPreemptionW import LgfsMultiServerPreemptionW



from numpy import infty
import numpy
from heapq import heappush, heappop
from aoi_simulator.queues.AoIQueueServer import AoIQueueServer
import queueing_tool as qt

"""
Extends QueueServer class from queueing_tool package.
Last Come First Served (LCFS) Queue with finite capacity and preemption.
Refrence: @ARTICLE{Costa_packet_management_2016,
            author={Costa, Maice and Codreanu, Marian and Ephremides, Anthony},
            journal={IEEE Transactions on Information Theory},
            title={On the Age of Information in Status Update Systems With Packet Management},
            year={2016},
            volume={62},
            number={4},
            pages={1897-1910},
            doi={10.1109/TIT.2016.2533395}}
Parameters
    ----------
    preemption : int (0: preemption in service, 1: preemption in waiting)
        Preemption in service: No waiting queue. New arrival replaces agent in service. (M/M/1/1*)
        Preemption in waiting: One waiting slot. New arrival replaces agent in waiting queue. (M/M/1/2*)
    **kwargs
        Any :class:`~QueueServer` parameters.
"""

class LcfsPreemption(AoIQueueServer):

    def __init__(self, preemption=0, **kwargs):
        super(LcfsPreemption, self).__init__(**kwargs)
        self.preemption = preemption

    def __repr__(self):
        tmp = ("LcfsPreemption:{0}. Servers: {1}, queued: {2}, arrivals: {3}, "
               "departures: {4}, next time: {5}")
        arg = (self.edge[2], self.num_servers, len(self.queue), self.num_arrivals,
               self.num_departures, round(self._time, 3))
        return tmp.format(*arg)

    def next_event(self):
        
        if self._departures[0]._time < self._arrivals[0]._time:
            new_depart = heappop(self._departures)
            self._current_t = new_depart._time
            self._num_total -= 1
            self.num_system -= 1
            self.num_departures += 1
            new_depart.queue_action(self, 2)

            if self.collect_data and new_depart.agent_id in self.data:
                self.data[new_depart.agent_id][-1][2] = self._current_t

            # Preemption in waiting 
            if self.preemption == 1 and len(self.queue) > 0:
                agent = self.queue.popleft()
                if self.collect_data and agent.agent_id in self.data:
                    self.data[agent.agent_id][-1][1] = self._current_t

                agent._time = self.service_f(self._current_t)
                agent.queue_action(self, 1)
                heappush(self._departures, agent)

            self._update_time()
            return new_depart

        elif self._arrivals[0]._time < infty:
            arrival = heappop(self._arrivals)
            self._current_t = arrival._time

            if self._active:
                self._add_arrival()

            self.num_system += 1
            self._num_arrivals += 1

            if self.collect_data:
                b = 0 if self.num_system <= self.num_servers else 1
                if arrival.agent_id not in self.data:
                    self.data[arrival.agent_id] = \
                        [[arrival._time, 0, 0, 0, self.num_system]]
                else:
                    self.data[arrival.agent_id]\
                        .append([arrival._time, 0, 0, 0, self.num_system])

            arrival.queue_action(self, 0)

            # Preemption in service
            if self.preemption == 0:
                # No agent in service
                if self.num_system == 1:
                    if self.collect_data:
                        self.data[arrival.agent_id][-1][1] = arrival._time
                    # Go directly to service
                    heappush(self._departures, arrival)
                # Agent already in service is replaced
                else:
                    # Remove agent in service
                    agent_replaced = heappop(self._departures)
                    self.num_system -= 1
                    # Include new agent in service
                    heappush(self._departures, arrival)
                    if self.collect_data:
                        self.data[arrival.agent_id][-1][1] = arrival._time
                        self.data[agent_replaced.agent_id][-1][1] = 0

                arrival.queue_action(self, 1)
                arrival._time = self.service_f(arrival._time)
                self._update_time()
            
            # Preemption in waiting
            elif self.preemption == 1:
                # No agent in service / waiting
                if self.num_system == 1:
                    if self.collect_data:
                        self.data[arrival.agent_id][-1][1] = arrival._time
                    # Go directly to service
                    heappush(self._departures, arrival)
                    arrival.queue_action(self, 1)
                    arrival._time = self.service_f(arrival._time)
                # One agent in service
                elif self.num_system == 2:
                    self.queue.append(arrival)
                    if len(self.queue) > 1:
                        print("Error: queue can't be bigger than 1 !")

                # One agent in service and one agent in waiting
                else:
                    # Remove agent in waiting
                    agent_replaced = self.queue.popleft()
                    self.num_system -= 1
                    # Include new arrival in waiting
                    self.queue.append(arrival)
                self._update_time()


from numpy import infty
import numpy
from heapq import heappush, heappop
import queueing_tool as qt
from aoi_simulator.queues.AoIQueueServer import AoIQueueServer
from reliability.Distributions import Weibull_Distribution as WB
import json 

"""
Extends QueueServer class from queueing_tool package.
Last Generated First Served (LGFS) Queue with preemption.

Parameters
    ----------
    preemption : int (0: preemption in service, 1: preemption in waiting, 2: preemption conditional)
        Preemption in service: No waiting queue. New arrival replaces agent in service.
        Preemption in waiting: One waiting slot. New arrival replaces agent in waiting queue.
        Preemption conditional: New arrival can preemp in service or in waiting, dinamically
    mrl_file: path to json file with mean residual life function, only for Preemption Conditional (examples in data_aux folder)
    **kwargs
        Any :class:`~QueueServer` parameters.
"""

class LgfsPreemption(AoIQueueServer):

    def __init__(self, preemption=0, mrl_file=None, **kwargs):
        super(LgfsPreemption, self).__init__(**kwargs)
        self.preemption = preemption
        with open(mrl_file, 'r') as f:
            self.MRL = json.load(f)
        self.last_departure_gen_time = 0
        self.contador = 0

    def __repr__(self):
        tmp = ("LcfsPreemption:{0}. Servers: {1}, queued: {2}, arrivals: {3}, "
               "departures: {4}, next time: {5}")
        arg = (self.edge[2], self.num_servers, len(self.queue), self.num_arrivals,
               self.num_departures, round(self._time, 3))
        return tmp.format(*arg)

    # Returns mean residual life for a packet with elapsed service time 't'
    # Approximate to the nearest time 't'
    def expected_sv(self, t):
            return self.MRL.get(t, self.MRL[min(self.MRL.keys(), key=lambda k: abs(float(k)-t))])

    def next_event(self):
        
        if self._departures[0]._time < self._arrivals[0]._time:
            new_depart = heappop(self._departures)
            self.last_departure_gen_time = new_depart.gen_time
            self._current_t = new_depart._time
            self._num_total -= 1
            self.num_system -= 1
            self.num_departures += 1
            new_depart.queue_action(self, 2)

            if self.collect_data and new_depart.agent_id in self.data:
                self.data[new_depart.agent_id][-1][2] = self._current_t

            # Preemption in waiting 
            if len(self.queue) > 0:
                agent = self.queue.popleft()
                if self.collect_data and agent.agent_id in self.data:
                    self.data[agent.agent_id][-1][1] = self._current_t
                agent._time = self.service_f(self._current_t)
                agent.queue_action(self, 1)
                heappush(self._departures, agent)

            self._update_time()
            return new_depart

        elif self._arrivals[0]._time < infty:
            arrival = heappop(self._arrivals)
            self._current_t = arrival._time

            if self._active:
                self._add_arrival()

            self.num_system += 1
            self._num_arrivals += 1

            if self.collect_data:
                b = 0 if self.num_system <= self.num_servers else 1
                if arrival.agent_id not in self.data:
                    self.data[arrival.agent_id] = \
                        [[arrival._time, 0, 0, 0, self.num_system]]
                else:
                    self.data[arrival.agent_id]\
                        .append([arrival._time, 0, 0, 0, self.num_system])

            arrival.queue_action(self, 0)

            # Preemption in service
            if self.preemption == 0:
                # No agent in service
                if self.num_system == 1:
                    if self.collect_data:
                        self.data[arrival.agent_id][-1][1] = arrival._time
                    # Go directly to service
                    heappush(self._departures, arrival)
                    arrival.queue_action(self, 1)
                    arrival._time = self.service_f(arrival._time)

                # Agent already in service is replaced if new arrival is fresher
                else:
                    agent_in_sv = self._departures[0]
                    if arrival.gen_time > agent_in_sv.gen_time:
                        # Remove agent in service
                        agent_replaced = heappop(self._departures)
                        # Include new agent in service
                        heappush(self._departures, arrival)
                        if self.collect_data:
                            self.data[arrival.agent_id][-1][1] = arrival._time
                            self.data[agent_replaced.agent_id][-1][1] = 0
                        arrival.queue_action(self, 1)
                        arrival._time = self.service_f(arrival._time)
                    self.num_system -= 1
                self._update_time()
            
            # Preemption in waiting
            elif self.preemption == 1:
                # No agent in service / waiting
                if self.num_system == 1:
                    if self.collect_data:
                        self.data[arrival.agent_id][-1][1] = arrival._time
                    # Go directly to service
                    heappush(self._departures, arrival)
                    arrival.queue_action(self, 1)
                    arrival._time = self.service_f(arrival._time)

                # One agent in service
                elif self.num_system == 2:
                    self.queue.append(arrival)
                    if len(self.queue) > 1:
                        print("Error: queue can't be bigger than 1 !")

                # One agent in service and one agent in waiting
                else:
                    agent_in_queue = self.queue[0]
                    if arrival.gen_time > agent_in_queue.gen_time:
                        # Remove agent in waiting
                        agent_replaced = self.queue.popleft()
                        # Include new arrival in waiting
                        self.queue.append(arrival)             
                    self.num_system -= 1
                self._update_time()

            # Preemption conditional
            elif self.preemption == 2:
                # No agent in service
                if self.num_system == 1:
                    if self.collect_data:
                        self.data[arrival.agent_id][-1][1] = arrival._time
                    # Go directly to service
                    heappush(self._departures, arrival)
                    arrival.queue_action(self, 1)
                    arrival._time = self.service_f(arrival._time)
                # One agent in service
                elif self.num_system == 2:
                    agent_in_sv = self._departures[0]
                    if arrival.gen_time > agent_in_sv.gen_time:
                        # New agent is fresher
                        s0 = self.last_departure_gen_time
                        s1 = agent_in_sv.gen_time
                        s2 = arrival.gen_time
                        MRL = self.expected_sv(self._current_t - self.data[agent_in_sv.agent_id][-1][1])
                        if ((s1 - s0)*1.384) < ((s2 - s0)*MRL):
                        # Remove agent in service  
                            agent_replaced = heappop(self._departures)
                            self.num_system -= 1
                            # Include new agent in service
                            heappush(self._departures, arrival)
                            if self.collect_data:
                                self.data[arrival.agent_id][-1][1] = arrival._time
                                self.data[agent_replaced.agent_id][-1][1] = 0
                            arrival.queue_action(self, 1)
                            arrival._time = self.service_f(arrival._time)
                        else:
                        # New arrival is queued
                            self.queue.append(arrival)             
                            self.contador += 1    
                    else:
                        self.num_system -= 1
                # One agent in service and one agent in waiting     
                else:    
                    agent_in_queue = self.queue[0]
                    agent_in_sv = self._departures[0]
                    if arrival.gen_time > agent_in_queue.gen_time:
                        # Remove agent in queue (obsolete)
                        agent_replaced = self.queue.popleft()
                        s0 = self.last_departure_gen_time
                        s1 = agent_in_sv.gen_time
                        s2 = arrival.gen_time
                        MRL = self.expected_sv(self._current_t - self.data[agent_in_sv.agent_id][-1][1])
                        if ((s1 - s0)*1.384) < ((s2 - s0)*MRL):
                        # Remove agent in service  
                            agent_replaced = heappop(self._departures)
                            self.num_system -= 1
                            # Include new agent in service
                            heappush(self._departures, arrival)
                            if self.collect_data:
                                self.data[arrival.agent_id][-1][1] = arrival._time
                                self.data[agent_replaced.agent_id][-1][1] = 0
                            arrival.queue_action(self, 1)
                            arrival._time = self.service_f(arrival._time)
                        else:
                        # New arrival is queued
                            self.queue.append(arrival)             
                            self.contador += 1   
                    self.num_system -= 1
                self._update_time()


import queueing_tool as qt
from aoi_simulator.core.AoIAgent import AoIAgent
from heapq import heappush
import numpy as np

"""
Extends QueueServer class from queueing-tool package.
Basic First Come First Served (FCFS) Queue for AoI simulation. 
Reference: @INPROCEEDINGS{Yates_often,
            author={Kaul, Sanjit and Yates, Roy and Gruteser, Marco},
            booktitle={2012 Proceedings IEEE INFOCOM},
            title={Real-time status: How often should one update?},
            year={2012},
            pages={2731-2735},
            doi={10.1109/INFCOM.2012.6195689}}
Parameters
    ----------
    **kwargs
        Any :class:`~QueueServer` parameters.
"""

class AoIQueueServer(qt.QueueServer):

    def __init__(self, **kwargs):
        super(AoIQueueServer, self).__init__(**kwargs)
        self.AgentFactory = AoIAgent

    def __repr__(self):
        tmp = ("AoiQueueServer:{0}. Servers: {1}, queued: {2}, arrivals: {3}, "
               "departures: {4}, next time: {5}")
        arg = (self.edge[2], self.num_servers, len(self.queue), self.num_arrivals,
               self.num_departures, round(self._time, 3))
        return tmp.format(*arg)

    # Overrides method to include agent generated time
    def _add_arrival(self, agent=None):
        if agent is not None:
            self._num_total += 1
            heappush(self._arrivals, agent)
        else:
            if self._current_t >= self._next_ct:
                self._next_ct = self.arrival_f(self._current_t)

                if self._next_ct >= self.deactive_t:
                    self._active = False
                    return

                self._num_total += 1
                new_agent = self.AgentFactory((self.edge[2], self._oArrivals), self._next_ct)
                new_agent._time = self._next_ct
                new_agent.gen_time = self._next_ct # agent generated time
                heappush(self._arrivals, new_agent)

                self._oArrivals += 1

                if self._oArrivals >= self.active_cap:
                    self._active = False

        if self._arrivals[0]._time < self._departures[0]._time:
            self._time = self._arrivals[0]._time


import json
import numpy as np
from scipy.signal import correlate

"""
Base class that calculates Age of Information (AoI)

"""

class AoICalc():
    
    def __init__(self, sim_id, data_file, num_sources, save_AoI_seq="None", save_Q_seq="None", **kwargs):

        self.sim_id = sim_id
        self.data_file = data_file
        self.num_sources = num_sources
        self.save_AoI_seq = save_AoI_seq
        self.save_Q_seq = save_Q_seq
        
        self.aoi = {}

        for i in range(self.num_sources):
            self.aoi[i] = {}
            self.aoi[i]["Preempted"] = 0
            self.aoi[i]["Obsolete"] = 0
            self.aoi[i]["RMSE"] = 0

        # Load data
        with open (self.data_file, 'r') as d:
            data = json.load(d)

        # Split sources
        self.splitted_data = {}
        for id, value in data.items():
            id_splitted = id.split(",")
            source = int(id_splitted[0][1:])
            gen_time = float(id_splitted[2][0:-1])
            deliv_time = value[2]
            times = (gen_time, deliv_time)
            if source in self.splitted_data:
                # Remove preempted
                if deliv_time == 0:
                    self.aoi[source]["Preempted"] += 1
                    continue
                # Remove obsolete
                elif gen_time < self.splitted_data[source][-1][0]:
                    self.aoi[source]["Obsolete"] += 1
                    continue
                self.splitted_data[source].append(times)
            else:
                self.splitted_data[source] = []
                self.splitted_data[source].append(times)

    def save_seqences(self, i):
        if self.save_AoI_seq != "None":
            try:
                arq_name = self.save_AoI_seq + "/" + self.sim_id + "_source_" + str(i) + ".txt"
                with open(arq_name, 'w') as f:
                    f.write(str(self.Q_vector))
            except Exception as e:
                print("Error saving AoI sequence: %s" %e)

        if self.save_Q_seq != "None":
            try:
                arq_name = self.save_Q_seq + "/" + self.sim_id + "_source_" + str(i) + ".txt"
                with open(arq_name, 'w') as f:
                    f.write(str(self.Q_vector))
            except Exception as e:
                print("Error saving Q sequence: %s" %e)

"""
Class that calculates Mean Age of Information (Mean AoI)

Parameters
----------
sim_id: str
    identification of simulation
data_file : json file
    json with class:Agent "agent_id" as keys and list [arrival, service
    start, departure times, agents waiting to be served and the total agents in the
    system] as values
num_sources: int
    number of sources of Agents. Mean AoI will be calculated for each source - Integer
save_AoI_seq: str (optional, defalut: "None")
    path to folder where to save the sequence of calculated AoI
    example: "/tmp/my_sim"
save_Q_seq: str (optional, defalut: "None")
    path to folder where to save the sequence of calculated AoI areas Q
    example: "/tmp/my_sim"

Returns
----------
mean_aoi: dict
    dict with sources as keys and {"MeanAoI", "RMSE", "Preempted", "Obsolete"} as values
"""
class MeanAoICalc(AoICalc):

    def __init__(self, sim_id, data_file, num_sources, save_AoI_seq="None", save_Q_seq="None", **kwargs):
        super().__init__(sim_id, data_file, num_sources, save_AoI_seq, save_Q_seq, **kwargs)

        self.aoi_vector = []
        self.Q_vector = []
        for i in range(self.num_sources):
            self.aoi[i]["MeanAoI"] = np.inf

        print("Calculating Mean AoI values for simulation [%s]" %self.sim_id)

        for i in self.splitted_data.keys():
            try:
                self.calc_aoi(i)
            except Exception as e:
                print("Error calculating AoI for source %d: %s" %(i, e))
            try:
                self.save_seqences(i)
            except Exception as e:
                print("Error saving sequences for source %d: %s" %(i, e))

        # Calculates mean AoI for a data array  
    def calc_aoi(self, source):
        
        self.aoi_vector.clear()
        self.Q_vector.clear()
        
        def Qi(Ti, Yi):
            Q = Ti*Yi + 0.5*(Yi**2)
            return Q
        
        data = self.splitted_data[source]
        ti = data[0][0] # Gen time packet #0
        t_start = ti
        ti_linha = 0
        Ti = 0
        Qi_total = 0
        N = 0 # Total delivered packets
        data.pop(0) # Remove first entry
        for value in data:
            if value[1] == 0:
                continue
            N = N + 1
            if value[0] < ti:
                print("Error: obsolete packet!")
                print("gen_time = %f" %value[0])
                break
            Yi = value[0] - ti
            ti = value[0]
            ti_linha = value[1]
            Ti = ti_linha - ti
            Qi_now = Qi(Ti, Yi)
            Qi_total = Qi_total + Qi_now
            mean_aoi = Qi_total/(ti_linha-t_start)
            self.aoi_vector.append(mean_aoi)
            self.Q_vector.append(Qi_now)
        Qi_total = Qi_total + 0.5*(Ti**2)
        self.aoi[source]["MeanAoI"] = Qi_total / (ti_linha - t_start)
        self.aoi[source]["RMSE"] = self.calc_RMSE(N, ti_linha)

    # Caclulate root mean squared error
    def calc_RMSE(self, N, tau):
        Q = self.Q_vector
        mean_q = np.mean(Q)
        var_q = np.var(Q)
        # Normalized Q
        norm_Q = (Q - mean_q)
        # Autocorrelation
        result = correlate(norm_Q, norm_Q, mode='same')
        acorr = result [N//2 + 1:]  / (var_q * np.arange(N-1, N//2, -1))
        M = len(acorr)
        iat = 0
        # IAT - Integrated autocorrelation times
        for i in range(M):
            iat = iat + (1 - (i+1)/M)*acorr[i]
        iat = 1 + 2*iat
        var_Q_mean = var_q/len(Q)
        true_var_Q_mean = var_Q_mean * iat
        var_aoi_mean = (N/tau)**2 * true_var_Q_mean
        RMSE = var_aoi_mean**0.5
        return RMSE

"""
Class that calculates Mean Peak Age of Information (Mean pAoI)

Parameters
----------
sim_id: str
    identification of simulation
data_file : json file
    json with class:Agent "agent_id" as keys and list [arrival, service
    start, departure times, agents waiting to be served and the total agents in the
    system] as values
num_sources: int
    number of sources of Agents. Mean AoI will be calculated for each source - Integer
save_AoI_seq: str (optional, defalut: "None")
    path to folder where to save the sequence of calculated AoI
    example: "/tmp/my_sim"
save_Q_seq: str (optional, defalut: "None")
    path to folder where to save the sequence of calculated AoI areas Q
    example: "/tmp/my_sim"

Returns
----------
mean_peak_aoi: dict
    dict with sources as keys and {"MeanPeakAoI", "RMSE", "Preempted", "Obsolete"} as values
"""
class MeanPeakAoICalc(AoICalc):

    def __init__(self, sim_id, data_file, num_sources, save_AoI_seq="None", save_Q_seq="None", **kwargs):
        super().__init__(sim_id, data_file, num_sources, save_AoI_seq, save_Q_seq, **kwargs)

        self.paoi_vector = []
        for i in range(self.num_sources):
            self.aoi[i]["MeanPeakAoI"] = np.inf

        print("Calculating Mean Peak AoI values for simulation [%s]" %self.sim_id)
                
        for i in self.splitted_data.keys():
            try:
                self.calc_aoi(i)
            except Exception as e:
                print("Error calculating AoI for source %d: %s" %(i, e))
            try:
                self.save_seqences(i)
            except Exception as e:
                print("Error saving sequences for source %d: %s" %(i, e))

    def calc_aoi(self, source):
        self.paoi_vector.clear()
        
        # peak age calculation
        def Ai(Ti, Yi):
            return Ti + Yi

        data = self.splitted_data[source]
        ti = data[0][0] # Gen time packet #0
        N = 0 # total delivered packets
        data.pop(0) # Remove first entry
        for value in data:
            if value[1] == 0: 
                continue
            N = N + 1 
            if value[0] < ti:
                print("Error: obsolete packet!")
                print("gen_time = %f" %value[0])
                break
            Yi = value[0] - ti
            ti = value[0]
            ti_linha = value[1]
            Ti = ti_linha - ti
            self.paoi_vector.append(Ai(Ti, Yi))
        self.aoi[source]["MeanPeakAoI"] = np.mean(self.paoi_vector)
        self.aoi[source]["RMSE"] = self.calc_RMSE(N, source)

    def calc_RMSE(self, N, source):
        # Variance
        var_A = np.var(self.paoi_vector)
        # Normalized A
        norm_A = (self.paoi_vector - self.aoi[source]["MeanPeakAoI"])
        # Autocorrelation
        result = correlate(norm_A, norm_A, mode='same')
        acorr = result [N//2 + 1:]  / (var_A * np.arange(N-1, N//2, -1))
        M = len(acorr)
        iat = 0
        # IAT
        for i in range(M):
            iat = iat + (1 - (i+1)/M)*acorr[i]
        iat = 1 + 2*iat
        var_A_mean = var_A/N
        true_var_A_mean = var_A_mean * iat
        RMSE = true_var_A_mean**0.5
        return RMSE



import queueing_tool as qt
import numpy as np
import collections
import numbers

"""
Extends QueueNetwork Class from queueing_tool package
- Includes a method to fetch data for AoI calculation

Parameters
    ----------
    **kwargs
        Any :class:`~QueueNetwork` parameters.
"""

class AoIQueueNetwork(qt.QueueNetwork):


    def __init__(self, **kwargs):
        super(AoIQueueNetwork, self).__init__(**kwargs)

    def __repr__(self):
        the_string = 'AoiQueueNetwork. # nodes: {0}, edges: {1}, agents: {2}'
        return the_string.format(self.nV, self.nE, np.sum(self.num_agents))

    def get_AoI_data(self, monitor=None):
        """Gets data from queues related to AoI.

        Parameters
        ----------
        monitor : int
            The edge index identifying the monitor where updates are delivered.

        Returns
        -------
        dict
            Returns a ``dict`` where the keys are the
            :class:`Agent's<.Agent>` ``agent_id`` and the values are
            :class:`ndarrays<~numpy.ndarray>` for that
            :class:`Agent's<.Agent>` data. The columns of this array
            are as follows:

            * First: The generation time of an agent (update) at his source.
            * Second: The start of service time of an agent (update) at the monitor.
            * Third: The departure time of an agent (update) at the monitor.
        """

        queues = _get_queues(self.g, monitor, None, None)
        monitor_queue = queues[0]
        data = {}
        for agent_id, dat in self.edge2queue[monitor_queue].data.items():
            datum = dat[0][0:3]
            if agent_id in data:
                data[agent_id] = np.vstack((data[agent_id], datum))
            else:
                data[agent_id] = datum
        return data

def _get_queues(g, queues, edge, edge_type):
    """Used to specify edge indices from different types of arguments."""
    INT = numbers.Integral
    if isinstance(queues, INT):
        queues = [queues]

    elif queues is None:
        if edge is not None:
            if isinstance(edge, tuple):
                if isinstance(edge[0], INT) and isinstance(edge[1], INT):
                    queues = [g.edge_index[edge]]
            elif isinstance(edge[0], collections.Iterable):
                if np.array([len(e) == 2 for e in edge]).all():
                    queues = [g.edge_index[e] for e in edge]
            else:
                queues = [g.edge_index[edge]]
        elif edge_type is not None:
            if isinstance(edge_type, collections.Iterable):
                edge_type = set(edge_type)
            else:
                edge_type = set([edge_type])
            tmp = []
            for e in g.edges():
                if g.ep(e, 'edge_type') in edge_type:
                    tmp.append(g.edge_index[e])

            queues = np.array(tmp, int)

        if queues is None:
            queues = range(g.number_of_edges())

    return queues


from numpy import infty
from heapq import heappush, heappop
from aoi_simulator.queues.AoIQueueServer import AoIQueueServer

"""
Extends QueueServer Class from queueing_tool package
Multi source and multi server model
Last Generated First Served Queue - Multi Preemption in Waiting
LGFS: Last generated packet is served first among all packets in queue
Multi Preemption in Waiting: Each source has a waiting queue, where preemption is allowed
Reference: not published yet...
Parameters
    ----------
    **kwargs
        Any :class:`~QueueServer` parameters.
"""

class LgfsMultiServerPreemptionW(AoIQueueServer):

    def __init__(self, **kwargs):
        super(LgfsMultiServerPreemptionW, self).__init__(**kwargs)
        # Dic with source index as keys and last departure generation time as values
        self._last_departures_gen_time = {"freshest": infty}
        # We define a queue as a dic with a key to each source
        self.multi_queue = {}

    def __repr__(self):
        tmp = ("LgfsMultiServerPreemptionW:{0}. Servers: {1}, queued: {2}, arrivals: {3}, "
               "departures: {4}, next time: {5}")
        arg = (self.edge[2], self.num_servers, len(self.queue), self.num_arrivals,
               self.num_departures, round(self._time, 3))
        return tmp.format(*arg)


    def enqueue(self, agent):
        source = agent.agent_id[0]
        if source in self.multi_queue.keys():
            gen_time = agent.gen_time
            if gen_time > self.multi_queue[source].gen_time:
                self.multi_queue[source] = agent
            self.num_system -= 1
        else:
            self.multi_queue[source] = agent

    # Getters and setters for last_departures
    def get_last_departures_gen_time(self, source):
        if source in self._last_departures_gen_time:
            return self._last_departures_gen_time[source]
        else:
            return 0

    def set_last_departures_gen_time(self, source, time):
        if source in self._last_departures_gen_time:
            if time < self._last_departures_gen_time[source]:
                return
        self._last_departures_gen_time[source] = time

    # # #
    
    def next_event(self):
    
        # Check if source has an agent in service
        def check_source_service(source):
            for agent in self._departures:
                if repr(agent) == 'InftyAgent':
                    continue
                if agent.agent_id[0] == source:
                    return True
            return False

        # # #

        # Return next agent to be served
        def fetch_queue():
            agent_served = None
            source_served = None
            # Max aged source in queue
            oldest_source = "freshest"
            for source, agent in self.multi_queue.items():
                # Source is already in service
                if check_source_service(source):
                    continue
                if self.get_last_departures_gen_time(source) < self.get_last_departures_gen_time(oldest_source):
                    oldest_source = source
                    agent_served = agent
                    source_served = source
            if agent_served:
                del self.multi_queue[source_served]
            return agent_served

        # # #

        # Departure
        if self._departures[0]._time < self._arrivals[0]._time:
            new_depart = heappop(self._departures)
            self._current_t = new_depart._time
            self._num_total -= 1
            self.num_system -= 1
            self.num_departures += 1
            new_depart_source = new_depart.agent_id[0]
            # Update last departure time
            self.set_last_departures_gen_time(new_depart_source, new_depart.gen_time)

            if self.collect_data and new_depart.agent_id in self.data:
                self.data[new_depart.agent_id][-1][2] = self._current_t

            # Fetch next agent according to the policy
            if bool(self.multi_queue):
                agent = fetch_queue()
                if agent is not None:
                    if self.collect_data and agent.agent_id in self.data:
                        self.data[agent.agent_id][-1][1] = self._current_t
                    agent._time = self.service_f(self._current_t)
                    agent.queue_action(self, 1)
                    heappush(self._departures, agent)

            new_depart.queue_action(self, 2)
            self._update_time()
            return new_depart

        elif self._arrivals[0]._time < infty:
            arrival = heappop(self._arrivals)
            self._current_t = arrival._time

            if self._active:
                self._add_arrival()

            self.num_system += 1
            self._num_arrivals += 1

            if self.collect_data:
                b = 0 if self.num_system <= self.num_servers else 1
                if arrival.agent_id not in self.data:
                    self.data[arrival.agent_id] = \
                        [[arrival._time, 0, 0, len(self.queue) + b, self.num_system]]
                else:
                    self.data[arrival.agent_id]\
                        .append([arrival._time, 0, 0, len(self.queue) + b, self.num_system])
            arrival.queue_action(self, 0)

            # We have idle server(s)
            if len(self._departures) <= self.num_servers:
                if self.collect_data:
                    self.data[arrival.agent_id][-1][1] = arrival._time
                arrival._time = self.service_f(arrival._time)
                arrival.queue_action(self, 1)
                heappush(self._departures, arrival)
            else:
                self.enqueue(arrival)

            self._update_time()


from queueing_tool.queues.agents import Agent

"""
Extends Agent Class from queueing_tool package
- Includes generated time in agent attributes

Parameters
    ----------
    **kwargs
        Any :class:`~Agent` parameters.
"""

class AoIAgent(Agent):

    def __init__(self, agent_id=(0, 0), gen_time=0, **kwargs):
        new_agent_id = (agent_id[0], agent_id[1], gen_time)
        super(AoIAgent, self).__init__(new_agent_id, **kwargs)
        self.gen_time = 0
   
    def __repr__(self):
        return "AoIAgent; agent_id:{0}. time: {1}. gen_time: {2}".format(self.agent_id, round(self._time, 3), round(self.gen_time,3))



from numpy import infty
from heapq import heappush, heappop
from aoi_simulator.queues.AoIQueueServer import AoIQueueServer

"""
Extends QueueServer Class from queueing_tool package
Multi source and multi server model
Last Generated First Served Queue with no preemption allowed
Policies:
    - LGFS: Last generated packet is served first among all packets in queue
    - MAX-LGFS: Last generated packet from the source with the maximum age is served first among all packets in queue
    - MASIF-LGFS: Last generated packet from the source with the maximum Age of Served Information (AoSI) is served first among all packets in queue
Reference: @misc{sun2018ageoptimal,
            title={Age-Optimal Updates of Multiple Information Flows}, 
            author={Yin Sun and Elif Uysal-Biyikoglu and Sastry Kompella},
            year={2018},
            eprint={1801.02394},
            archivePrefix={arXiv},
            primaryClass={cs.IT}
        }
Parameters
    ----------
    policy : int [0: LGFS (Last generated First Served), 1: MAX-LGFS (Maximum Age First), 2: MASIF-LGFS (Maximum Age of Served Information First)]
    **kwargs
        Any :class:`~QueueServer` parameters.
"""

class LgfsMultiServerNoPreemption(AoIQueueServer):

    def __init__(self, policy=0, **kwargs):
        super(LgfsMultiServerNoPreemption, self).__init__(**kwargs)
        self.policy = policy
        # Dic with source index as keys and last departure generation time as values
        self._last_departures_gen_time = {"oldest": infty}
        # Dic with source index as keys and last in service agent generation time as values
        self._in_service_gen_time = {"oldest": infty}

    def __repr__(self):
        tmp = ("LgfsMultiServerNoPreemption:{0}. Servers: {1}, queued: {2}, arrivals: {3}, "
               "departures: {4}, next time: {5}")
        arg = (self.edge[2], self.num_servers, len(self.queue), self.num_arrivals,
               self.num_departures, round(self._time, 3))
        return tmp.format(*arg)

    # Getters and setters

    def get_last_departures_gen_time(self, source):
        if source in self._last_departures_gen_time:
            return self._last_departures_gen_time[source]
        else:
            return 0

    def get_in_service_gen_time(self, source):
        if source in self._in_service_gen_time:
            return self._in_service_gen_time[source]
        else:
            return 0

    def set_last_departures_gen_time(self, source, time):
        if source in self._last_departures_gen_time:
            if time < self._last_departures_gen_time[source]:
                return
        self._last_departures_gen_time[source] = time

    def set_in_service_gen_time(self, source, time):
        if source in self._in_service_gen_time:
            if time < self._in_service_gen_time[source]:
                return
        self._in_service_gen_time[source] = time

    # # #
    
    def next_event(self):
    
        # Return next agent to be served
        def fetch_queue():
            agent_served = None

            # LGFS
            if self.policy == 0:
                newest_gen_time = 0
                # Get newest agent from all sources
                for agent in self.queue:
                    if repr(agent) == 'InftyAgent':
                        continue
                    if agent.gen_time > newest_gen_time:
                        newest_gen_time = agent.gen_time
                        agent_served = agent

            # MAX-LGFS
            elif self.policy == 1: 
                # Max age source
                oldest_source = "oldest"
                # Newest gen time from max age source
                newest_gen_time = 0
                for agent in self.queue:
                    if repr(agent) == 'InftyAgent':
                        continue
                    agent_source = agent.agent_id[0]
                    # If agents have same source, we choose the newest generated time
                    if agent_source == oldest_source:
                        if agent.gen_time > newest_gen_time:
                            agent_served = agent
                            newest_gen_time = agent.gen_time
                    elif self.get_last_departures_gen_time(agent_source) < self.get_last_departures_gen_time(oldest_source):
                        oldest_source = agent_source
                        newest_gen_time = agent.gen_time
                        agent_served = agent

            # MASIF-LGFS
            else:
                # Max age source
                oldest_source = "oldest"
                # Newest gen time from max age source
                newest_gen_time = 0
                for agent in self.queue:
                    if repr(agent) == 'InftyAgent':
                        continue
                    agent_source = agent.agent_id[0]
                    # If agents have same source, we choose the newest generated time
                    if agent_source == oldest_source:
                        if agent.gen_time > newest_gen_time:
                            agent_served = agent
                            newest_gen_time = agent.gen_time
                    elif self.get_in_service_gen_time(agent_source) < self.get_in_service_gen_time(oldest_source):
                        oldest_source = agent_source
                        newest_gen_time = agent.gen_time
                        agent_served = agent

            self.queue.remove(agent_served)
            return agent_served

        # # #

        if self._departures[0]._time < self._arrivals[0]._time:
            new_depart = heappop(self._departures)
            self._current_t = new_depart._time
            self._num_total -= 1
            self.num_system -= 1
            self.num_departures += 1
            new_depart_source = new_depart.agent_id[0]
            # Update last departure time
            self.set_last_departures_gen_time(new_depart_source, new_depart.gen_time)

            if self.collect_data and new_depart.agent_id in self.data:
                self.data[new_depart.agent_id][-1][2] = self._current_t

            # Fetch next agent according to the policy
            if len(self.queue) > 0:
                agent = fetch_queue()
                if self.collect_data and agent.agent_id in self.data:
                    self.data[agent.agent_id][-1][1] = self._current_t
                # Update in service time of agent
                self.set_in_service_gen_time(agent.agent_id[0], agent.gen_time)
                agent._time = self.service_f(self._current_t)
                agent.queue_action(self, 1)
                heappush(self._departures, agent)

            new_depart.queue_action(self, 2)
            self._update_time()
            return new_depart

        elif self._arrivals[0]._time < infty:
            arrival = heappop(self._arrivals)
            self._current_t = arrival._time
            arrival_source = arrival.agent_id[0]

            if self._active:
                self._add_arrival()

            self.num_system += 1
            self._num_arrivals += 1

            if self.collect_data:
                b = 0 if self.num_system <= self.num_servers else 1
                if arrival.agent_id not in self.data:
                    self.data[arrival.agent_id] = \
                        [[arrival._time, 0, 0, len(self.queue) + b, self.num_system]]
                else:
                    self.data[arrival.agent_id]\
                        .append([arrival._time, 0, 0, len(self.queue) + b, self.num_system])

            arrival.queue_action(self, 0)

            # We have idle server(s)
            if self.num_system <= self.num_servers:
                if self.collect_data:
                    self.data[arrival.agent_id][-1][1] = arrival._time
                # Update in service time
                self.set_in_service_gen_time(arrival_source, arrival.gen_time)
                arrival._time = self.service_f(arrival._time)
                arrival.queue_action(self, 1)
                heappush(self._departures, arrival)
                
            # All servers busy: agent is queued
            else:
                self.queue.append(arrival)

            self._update_time()